/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntityControl.cpp							*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/

#include "EntityControl.h"

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// 
//	Top level control structure for synth
//	
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

void EntityControl::Run() {

	Mode oldMode = SOUND;
	Mode newMode = SOUND;

	initialise();

	while(1) {

		switch(newMode) {
		 	
			case RETURN :		newMode = oldMode;
						break;

			case SOUND :		newMode = soundMode->enter();
						oldMode = SOUND;
						break;

			case EDIT :		newMode = editMode->enter(soundManager->getSound());
						oldMode = EDIT; 
						break;

			case STACK :		newMode = stackMode->enter(soundManager->getSound());
						break;
	
			case INPUT :		newMode = inputMode->enter();
						break;
					
			case OUTPUT :		newMode = outputMode->enter();
						break;
					
			case SAVE :		newMode = saveMode->enter(soundManager->getSound(), oldMode);
						break;

			case GLOBAL :		newMode = globalMode->enter();
						break;

			case SYSEX :		newMode = sysexMode->enter();
						break;

			case MIPS :		newMode = mipsMode->enter();
						break;
					
			case SAVE_GLOBAL :	newMode = saveGlobalMode->enter();
						break;

			default :		newMode = SOUND;
					
		}	
	}
}

void EntityControl::initialise() { 

	/* Create the message queues */

	createQueue("CTRQ", 128, 8);	// Control message queue (mainly panel input)
	createQueue("SYNQ", 256, 8);	// Synth message queue (midi, rtc, edit)
	createQueue("SSTQ", 16, 8);	// Sound stack (for loading sounds)
	createQueue("MIDQ", 1024, 8);	// Midi queue (for midi input)
	
	/* Non-dependent objects */
	
	paras = new ParaStructure;				// Object only
	flash = new EntityFlash;				// Object
	global = new EntityGlobal;				// Object only
	panel = new EntityPanel;				// Object & task

	/* Dependent objects */

	dsp = new EntityDsp(paras);				// Object only
	
	/* Midishare */
	midi = new EntityMidi(global);				// Object & task
	
	/* Create voice assigner to translate midi note events to dsp voice events */
	assign = new EntityVoiceAssigner(global, paras, dsp);	// Object only

	/* Create co-ordinator for midi, sounds and dsp */
	synth = new EntitySynth(global, paras, midi, dsp, assign); // Object & task

	/* Create managed pool of sound objects */
	soundManager = new EntitySoundManager(flash, paras); 	// Object & task

	/* Create UI Modes */
	saveMode = new EntitySaveMode(panel, flash);		 
	stackMode = new EntityStackMode(panel, global, paras);
	editMode = new EntityEditMode(panel, global, paras);		 
	soundMode = new EntitySoundMode(panel, soundManager, global);	
	sysexMode = new EntitySysexMode(panel, midi, flash, global, soundManager);
	globalMode = new EntityGlobalMode(panel, global, flash, sysexMode);
	saveGlobalMode = new EntitySaveGlobalMode(panel, flash, global);

	outputMode = new EntityOutputMode(panel, global);	// Output level meters
	inputMode = new EntityInputMode(panel, global);		// Input level meters
	mipsMode = new EntityMipsMode(panel, global);		// MIPS meter, and debug
	
	flash->setPriority(150);     	// Set priority for flash background write
	flash->load(global);		// Load global settings and version number
	
	/* Do fresh install on system with no files matching current version */
	checkFreshInstall();
	
	/* Create and start main tasks (these priority levels are critical) */
	
	synth->create("SYNP", 50);
	soundManager->create("SMNP", 50);	
	panel->create("PANP", 49);
	midi->create("MIDP", 50);

	soundManager->loadSound(0); 	// load sound A1

	//showPriorityLevels();		// List priority level of each task
	//ASSERT(0);			// Show task memory use
	//test();			// Miscellaneous tests
	
	TRACE("Initialisation complete\n");
}

void EntityControl::checkFreshInstall() {
	
	if(global->matchesVersion(	ENTITY_MODEL_NUMBER,
					ENTITY_MAJOR_VERSION_NUMBER, 
				  	ENTITY_MINOR_VERSION_NUMBER,
					ENTITY_TEST_VERSION_NUMBER)) return;

	/* Global file was either absent from flash, in which case it 
	 * contained dummy version values of -1, -1, and a fresh install 
	 * is required, or the global file was generated by a different 
	 * version, in which case a fresh install is also appropriate. */
			
 	TRACE("Fresh installation\n");
	
	global->loadDefaultValues();
	
	global->setVersion(	ENTITY_MODEL_NUMBER,
				ENTITY_MAJOR_VERSION_NUMBER, 
			  	ENTITY_MINOR_VERSION_NUMBER,
				ENTITY_TEST_VERSION_NUMBER);
		
	flash->save(global);

	/* Copy presets to user locations (ie bank X -> bank A) */

	unsigned char soundBuffer[ENTITY_SYSEX_MAX_MESSAGE_SIZE];

	int startSoundNumber = flash->getnWritableSounds();
	int endSoundNumber = flash->getnSounds();
	int firstDestination = 0;
	
	for(int src = startSoundNumber, dst = firstDestination; src < endSoundNumber; src++, dst++) {

		int bufSize = flash->load(src, soundBuffer); if(!bufSize) continue;
		flash->save(dst, soundBuffer, bufSize);
	}
}

char *EntityControl::getVersionStr() {

	static char vs[8];

	sprintf(vs, "  Version %i.%02i", ENTITY_MAJOR_VERSION_NUMBER, ENTITY_MINOR_VERSION_NUMBER);

	return vs;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
//	Debugging
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

void EntityControl::showPriorityLevels() {
	
	showPriority("IDLE");
	showPriority("SMNP");
	showPriority("CTRP");
	showPriority("TFLA");
	showPriority("MIPP");
	showPriority("SYNP");
	showPriority("PANP");
	showPriority("REP1");
	showPriority("REP2");
	showPriority("REP3");
	showPriority("REP4");
	showPriority("MIDP");
	showPriority("INTR");
}

void EntityControl::showPriority(char *taskName) {

	rtems_id 		rtid;
	rtems_status_code 	status;
	rtems_task_priority	new_priority = 0;
	rtems_task_priority	old_priority = 0;
       
	status = rtems_task_ident(rtemsName(taskName), RTEMS_SEARCH_ALL_NODES, &rtid);
	if(status != RTEMS_SUCCESSFUL) TRACE("ERROR: Cannot get task id for %s\n", taskName);

	rtems_task_set_priority(rtid, new_priority, &old_priority);
	rtems_task_set_priority(rtid, old_priority, &old_priority);
	
	TRACE("%s priority level = %i\n", taskName, old_priority);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
//	Testing
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

bool EntityControl::test() {

	TRACE("TEST: Starting software testing\n");

	bool t1 = testEntitySound();
	bool t2 = testEntityFlash();

	return t1 && t2;
}

bool EntityControl::testEntitySound() {

	// This is a non-exhaustive test of the EntitySound class.
	// It randomises a sound, creates a checksum, compresses 
	// the sound, decompresses it, creates a new checksum and
	// compares it with the original checksum to see if they 
	// match. It does this several times.

	EntitySound *snd = soundManager->getSound();
	
	char *buffer = new char [2048];
	bool okay = true;

	TRACE("TEST: Testing sound compression\n");

	for(int x = 0; x < 10; x++) {

		snd->loadRandomValues();

		unsigned int c = snd->checksum();

		int size = snd->compress(buffer);
		TRACE("Compressed data size = %i bytes. ", size);
	
		snd->deCompress(buffer);
		unsigned int d = snd->checksum();

		if(d == c) TRACE("Checksums match\n");
		else { TRACE("Bad checksum\n"); okay = false; }
	}

	if(!okay) TRACE("ERROR: checksums did not match after compression cycle\n");
	else	  TRACE("PASS: EntitySound class appears to be working correctly\n");
	
	return okay;
}

bool EntityControl::testEntityFlash() {

	// This is a fairly detailed set of tests for the flash file system
	
	EntitySound *snd = soundManager->getSound();
	
	bool okay = true;

	TRACE("TEST: Testing flash file system\n"); flash->report();
	
	for(int x = 0; x < FLASH_MAX_SOUNDS; x++) {

		snd->loadRandomValues(); snd->setNumber(x);
		
		unsigned int c = snd->checksum();

		if(flash->save(snd)) TRACE("Written sound file %i.\t", x);
		else { TRACE("ERROR: could not write sound file %i.\t", x); okay = false; }

		snd->clearSound();
		
		if(flash->load(snd, x)) TRACE("Read sound file %i.\t", x);
		else { TRACE("ERROR: could not read sound file %i.\t", x); okay = false; }

		unsigned int d = snd->checksum();

		if(d == c) TRACE("Checksums match.\n");
		else { TRACE("Bad checksum.\n"); okay = false; }
	}

	int z1 = flash->getnFiles(); 
	TRACE("Rescanning flash...\n"); flash->scan();
	int z2 = flash->getnFiles(); TRACE("Done rescan.\n");

	if(z1 == z2) TRACE("Flash state tracking (no. spare files etc) seems okay\n");
	else { TRACE("ERROR: Flash state tracking faulty\n"); okay = false; }
	
	TRACE("Number of flash files = %i\n", flash->getnFiles());
	TRACE("Writing additional versions of some sounds...\n");

	for(int y = 19; y > 0; y -= 2) {
		
		TRACE("Overwriting sound %i, %i times\n", y, y);
		
		for(int x = 0; x < y; x++) {

			snd->loadRandomValues(); snd->setNumber(y);

			unsigned int c = snd->checksum();

			if(flash->save(snd)) TRACE("Written sound file %i.\t", y);
			else { TRACE("ERROR: could not overwrite sound file %i.\t", y); okay = false; }

			snd->clearSound();	
			if(flash->load(snd, y)) TRACE("Read sound file %i.\t", y);
			else { TRACE("ERROR: could not read sound file %i.\t", y); okay = false; }

			unsigned int d = snd->checksum();

			if(d == c) TRACE("Checksums match.\n");
			else { TRACE("Bad checksum.\n"); okay = false; }
		}
	}

	z1 = flash->getnFiles(); TRACE("Rescanning flash...\n"); flash->scan();
	z2 = flash->getnFiles(); TRACE("Done rescan.\n");

	if(z1 == z2) TRACE("Flash state tracking (no. spare files etc) seems okay\n");
	else { TRACE("ERROR: Flash state tracking faulty\n"); okay = false; }

	TRACE("Number of flash files = %i\n", flash->getnFiles());
	TRACE("Accessing versions\n");

	for(int x = 0; x < 30; x++) {

		int v = flash->initVersionAccess(x);
		TRACE("Number of versions of sound %i = %i\n", x, v);

		for(int z = 0; z < v; z++) {

			if(flash->loadVersion(snd, z)) TRACE("Read version %i of sound file %i.\n",v-z, x);
			else { TRACE("ERROR: Could not read version %i of sound file %i.\n",v-z, x); okay = false; }
		}
	}

	if(!okay) { TRACE("ERROR: Flash tests failed\n"); return false; }
	
	TRACE("PASS: EntityFlash class appears to be working correctly.\n");
	
	return okay;	
}

