/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntityGlobal.h								*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/

#ifndef __ENTITYGLOBAL_H__
#define __ENTITYGLOBAL_H__

#include <string.h>
#include <stdio.h>
#include <rtems.h>

#include "EntityMessageHandler.h"
#include "EntityMessages.h"
#include "EntityStatus.h"
#include "EntityDefines.h"

#include "utils.h"

#define MAX_GLOBAL_PARAS 32
#define MAX_GLOBAL_NAME_CHARS 24

// Global parameters are unlike sound parameters, in that the Global class includes the 
// actual data values, not an index. This is because there is only ever one global object.

enum SaveType { SAVE_CONFIG = 0, MIDI_SAVE_SOUND, MIDI_SAVE_BANK, MIDI_SAVE_ALL, MIDI_SAVE_CONFIG };

class GlobalPara {

protected:

	static int 	nGlobalParas;	// For counting parameter instantiations

	char 		name[MAX_GLOBAL_NAME_CHARS];

	int		defval;
	int		value;
	int		engval;
	
	int		min;
	int		max;

	bool		useDefault;
	
public:

	GlobalPara(char*, int min_ = 0, int max_ = +100, int default_ = 0);

	virtual	int 	getValue() { return useDefault ? defval : value; };

	virtual void 	setValue(int x) { useDefault = false; value = validValue(x); };
	virtual bool 	setDefault() { useDefault = true; return true; };
	virtual bool 	clearDefault() { useDefault = false; return true; };
	virtual	void 	setValueToDefault() { value = defval; useDefault = false; };
	virtual int	translate() { return useDefault ? defval : value; };
	virtual	Status 	incValue(int i);

	virtual int	getmin() { return min; } ;
	virtual int	getmax() { return max; } ;
	
	virtual	char* 	getName() { return name; };
	
	virtual char	*getValStr(int &strLen);
	virtual int	validValue(int val_);

	virtual int	getAcceleration() { return 0; };
	static void 	countZero() { nGlobalParas = 0; };
	static int 	getCount() { return nGlobalParas; };

	virtual 	SaveType save() { return SAVE_CONFIG; };

	virtual	bool	transmit() { return false; };
	virtual	bool	receive() { return false; };
};

class GlobalTunePara : public GlobalPara, public EntityMessageHandler {

	rtems_id	synq;
	void		updateSynth();
	
public:
	GlobalTunePara(char *n, int = 0);

	virtual	int 	translate();
	virtual void 	setValue(int x);
	virtual char	*getValStr(int &strLen);
	virtual int	getAcceleration() { return 100; };
	virtual bool 	setDefault(); 
	virtual bool 	clearDefault(); 
	virtual	void 	setValueToDefault();
       	virtual	Status 	incValue(int);	
};
	
class GlobalCalibrateTuningPara : public GlobalTunePara {

public:
	GlobalCalibrateTuningPara(char *n, int = 0);

	virtual char	*getValStr(int &strLen);
};
	
class GlobalSavePara : public GlobalPara {

	SaveType	saveType;
	
public:		

	GlobalSavePara(char*, SaveType);

	SaveType	save() { return saveType; };	
	virtual char	*getValStr(int &strLen);
};

class GlobalStringPara : public GlobalPara {

protected:

	char		**paraString;
	int		*paraLength;
	int		addIndex;
	
	virtual void	addString(char*);

public:

	GlobalStringPara(char*, int, int = 0);
	virtual ~GlobalStringPara();
	
	virtual char	*getValStr(int &strLen);	
};

class GlobalCascadePara : public GlobalStringPara, public EntityMessageHandler {

	rtems_id	synq;
	void		updateSynth();

	struct {	signed char machineIndex; 
			signed char nMachines; 
	
	} lut[N_CASCADE_MODES+1];	 
	
public:

	GlobalCascadePara(char *n, int = 1);

	virtual void 	setValue(int x);
	virtual bool 	setDefault(); 
	virtual bool 	clearDefault(); 
	virtual	void 	setValueToDefault();
       	virtual	Status 	incValue(int);	
	
	inline int	getMachineIndex() { return (int)lut[value].machineIndex; }
	inline int	getnMachines() { return (int)lut[value].nMachines; }
};
	
class GlobalOnOffPara : public GlobalStringPara {

public:
	GlobalOnOffPara(char *n, int = 0);
};

class GlobalStereoPara : public GlobalStringPara {

public:
	GlobalStereoPara(char *n, int = 0);
};

class GlobalReceivePara : public GlobalStringPara {

public:
	GlobalReceivePara(char *n, int = 0);
};

class GlobalTChanPara : public GlobalStringPara {

public:
	GlobalTChanPara(char *n, int = 0);
};

class GlobalRChanPara : public GlobalStringPara {

public:
	GlobalRChanPara(char *n, int = 0);
};

class GlobalDevIdPara : public GlobalStringPara {

public:
	GlobalDevIdPara(char *n, int = 0);
	virtual int 	translate();
};

class GlobalTRPara : public GlobalStringPara {

public:
	GlobalTRPara(char *n, int = 0);
	
	virtual	bool	transmit();
	virtual bool	receive();
};

class GlobalMidiCtrPara : public GlobalStringPara {

protected:
	
	char		*ctrLUT;	// Points to EntityGlobal::ctrLUT for looking up controller mapping
	char		ctrBit;		// LUT bit for this controller (ie CTR1 = 0x02, CTR2 = 0x04...)

	int		*paraValue;
	virtual	void	addString(char*, int);

public:

	int 		translate() { return useDefault ? paraValue[defval] : paraValue[value]; }

	void 		setValue(int x);
	bool 		setDefault();
	bool 		clearDefault();
	void 		setValueToDefault();
	Status 		incValue(int i);
	void		bTrace(int);

	GlobalMidiCtrPara(char *n, int defVal_, char ctrBit_, char *ctrLUT_);
	virtual ~GlobalMidiCtrPara();
};

class GlobalMidiCtr7Para : public GlobalMidiCtrPara {

public:

	GlobalMidiCtr7Para(char *n, int defVal_, char ctrBit_, char *ctrLUT_);
};	

class GlobalMidiCtr14Para : public GlobalMidiCtrPara {

public:

	GlobalMidiCtr14Para(char *n, int defVal_, char ctrBit_, char *ctrLUT_);
};

class EntityGlobal {

protected:
	
	GlobalPara* 	globalPara[MAX_GLOBAL_PARAS];	// The array of global parameters
	char		potLUT[128];			// For fast para selection
	
	int		currentParameterIndex;
	int		nVisibleParas;			// No. directly accessible to user
	int		baseParaIndex;
	int		nParas;				// No. of global parameters
	
	// Pointers to some global parameters (avoids using a hash)

	GlobalPara	*midictr1, *midictr2, *midictr3, *midimodl;
	GlobalPara	*midictrl, *midiprgr, *midivolr, *midiedit;
	GlobalPara	*midiptch, *midiecho, *midirchn, *miditchn;
	GlobalPara	*glbltune, *calitune, *fastedit, *shiftkey;
	GlobalPara	*mididvid, *entitymn, *majorvrn, *minorvrn;
	GlobalPara	*testvrsn, *encrctr1;

	GlobalCascadePara *midicasc;
	
	char		ctrLUT[128];			// Controller mapping LUT

public:

	EntityGlobal();
	
	inline int 	getnParas() { return nParas; };
	inline int 	getnVisibleParas() { return nVisibleParas; };
	
	bool		parameterInc(int x);
	bool		selectFirstParameter();
	bool		selectLastParameter();
	int		getCurrentParameterIndex();
	int		setParameterByIndex(int x);
	bool		setParameterByPot(int p);
	GlobalPara*	getCurrentParameter();
	GlobalPara* 	getParameterByIndex(int index);	

	GlobalPara*	getCalibrateTuningPara();

	bool		unpack(unsigned char*);		// For loading global values
	int		pack(unsigned char*);		// For saving them
	void		loadDefaultValues();

	bool		matchesVersion(int entityModelNumber, int majorVersionNumber, int minorVersionNumberm, int testVersionNumber);
	void		setVersion(int entityModelNumber, int majorVersionNumber, int minorVersionNumber, int testVersionNumber);

	bool		midiReceive(int chan);		// returns true if channel is valid for receive
	bool		midiDeviceReceive(int devId);	// returns true if device ID is valid for receive

	bool		fastEdit();			// returns true if using pots for fast edits
	bool		toggleFastEdit();		// toggle using pots for fast parameter access
	bool		shiftKey();			// returns true if shift is active during editing
	bool		toggleShiftKey();		// toggle shift during editing
	
	inline int	midiDeviceId() { return mididvid->translate(); }; // 0-15 or 127

	inline int	midiModlCtrl() { return midimodl->translate(); }; // Returns controller number 
	inline int	midiCtr1Ctrl() { return midictr1->translate(); }; //		 "
	inline int	midiCtr2Ctrl() { return midictr2->translate(); }; //		 "
	inline int	midiCtr3Ctrl() { return midictr3->translate(); }; //		 "
	
	inline int	masterTuning() { return glbltune->translate() + calitune->translate(); };
	inline int	midiRxChan() { return midirchn->translate(); };
	inline int	midiTxChan() { return miditchn->translate(); };
	inline int	midiPitch() { return midiptch->translate(); };
	
	inline bool	midiEcho() { return midiecho->translate(); };
	inline bool	encrCtr1() { return encrctr1->translate(); };

	inline int	midiCascadeGetnMachines() { return midicasc->getnMachines(); };
	inline int	midiCascadeGetMachineIndex() { return midicasc->getMachineIndex(); };

	inline bool	midiCtrlTransmit() { return midictrl->transmit(); };
	inline bool	midiCtrlReceive() { return midictrl->receive(); };

	inline bool	midiProgTransmit() { return midiprgr->transmit(); };
	inline bool	midiProgReceive() { return midiprgr->receive(); };

	inline bool	midiVolTransmit() { return midivolr->transmit(); };
	inline bool	midiVolReceive() { return midivolr->receive(); };

	inline bool	midiEditTransmit() { return midiedit->transmit(); }; 
	inline bool	midiEditReceive() { return midiedit->receive(); }; 

	inline char	ctrLookup(int cn) { return ctrLUT[cn&0x7f]; };
};















#endif /* __ENTITYGLOBAL_H__ */
