/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntitySound.h								*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005                                      *
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 ********************************************************************************/

#ifndef __ENTITYSOUND_H__
#define __ENTITYSOUND_H__

#include <rtems.h>
#include <chameleon.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "utils.h"
#include "EntityStatus.h"
#include "EntityDefines.h"

// Make sure this enum matches matrix strings in the cpp file

enum { 	CTR1=0, CTR2,	CTR3,	MMOD,
	KFLW, 	KVEL, 	VDIF,	STDF,
	LFO1, 	LFO2,	LFO3, 	LFO4,	
	LFO5, 	LFO6,	ENV1, 	ENV2,	 };

//----------------------------------------------
// A base class for naming parameters and pages

class ParaName {

	char* 	name;
	char* 	shortName;

	char	fixedName[MAX_PARANAME_CHARS];
	
	char*	fixName(char*, int);		// For selecting various caps options
	char*	lcase(char*);
	char*	caps(char*);
	char*	initcap(char*);
	
	int	nameLength;

public:

	ParaName(char*, char*);

	char*	getName();			// Name used in parameter edit
	char*	getName(int);			// " with caps/lcase options
	char*	getShortName();			// Name used in matrix edit
	char*	getShortName(int);		// " with caps/lcase options
};

//----------------------------------------------------------------------------
//
// 				Parameter
//
//----------------------------------------------------------------------------

class Parameter : public ParaName {

protected:

	static	int 	nParameters;		// To count no. of Parameter objects in each page
	static	int 	nParametersTotal;	// To count total no. of Parameter objects
	
	int		paraNumber;		// Number of parameter 
	int 		paraIndex;		// To determine offset into array of int sound values (n per para)
	int		dspParaIndex;		// Same parameters are in different order in dsp memory (1 per para)

	int		nrpnPotScale;		// For low res NRPN data, use customised parameter scaling
	int		nrpnPotOffset;		// Offset used with above
	int		nrpnPotScaleDepth;	// Same again, for matrix depths
	int		nrpnPotOffsetDepth;
	
	void		nrpnSetPotMinMax(int pmin, int pmax);
	
	int		min;
	int		max;
	
public:

	Parameter(char*, char*, int rmin= -8000, int rmax = +7999);

	inline int	getParaNumber() { return paraNumber; };
	inline int 	getParaIndex() { return paraIndex; }; 
	inline int	getDspParaIndex() { return dspParaIndex; };
	inline void	setDspParaIndex(int i) { dspParaIndex = i; };

	virtual int	translateNRPNPotValue(int potVal, int source);
	
	virtual char	*getValStr(int value, int source, int &strLen);
	virtual char	*baseValStr(int value, int &strLen);
	virtual char	*valStr(int value, int &strLen);

	virtual int	validValue(int value, int source);
	virtual	int 	translate(int value, int source);
	virtual int	mTranslate(int depth);

	virtual int	getmin() { return min; } ;
	virtual int	getmax() { return max; } ;

	virtual int	getAcceleration() { return 100; };
	virtual bool	isMutable() { return true; };
	virtual bool	isDspParameter() { return true; };
	virtual bool	isMeteredParameter() { return false; };
	virtual bool	usesPoPsMatrix() { return false; };
	virtual	bool	isContinuous(int source) { return true; };
	virtual bool	isStoppedAt1000() { return true; };
	
	static void 	allCountZero() { nParametersTotal = 0; };	
	static int 	getCountTotal() { return nParametersTotal; };
	static void 	countZero() { nParameters = 0; };
	static int 	getCount() { return nParameters; };
};

class PosParameter : public Parameter {

public:
	
	PosParameter(char *n, char *sn, int min = 0, int max = 7999);
	virtual int 	translate(int value, int source);
};

class PhaseParameter : public Parameter {

public:	

	PhaseParameter(char *n, char *sn, int min = -8000, int max = 7999);
	
	virtual int 	translate(int v, int source);
	virtual int 	mTranslate(int depth);
};

class vLFORateParameter : public PosParameter {

public:
	
	vLFORateParameter(char *n, char *sn, int min = 0, int max = 7999);
};

class sLFORateParameter : public PosParameter {

public:
	
	sLFORateParameter(char *n, char *sn, int min = -24, int max = 7999);
	virtual int 	translate(int value, int source);
	virtual char	*baseValStr(int value, int &strLen);	
};

class OffsetParameter : public Parameter {

public:
	
	OffsetParameter(char *n, char *sn, int min = -8000, int max = 7999);

	virtual int 	translate(int value, int source);
	virtual char	*baseValStr(int value, int &strLen);
	virtual bool	isStoppedAt1000() { return false; };
};

class SwitchParameter : public Parameter {

protected:

	int	ip;	

public:
	SwitchParameter(char *n, char *sn, int min, int max);

	virtual int 	translate(int value, int source);
	virtual int	getAcceleration() { return 0; };
	virtual	bool	isContinuous(int source);
	virtual bool	isStoppedAt1000() { return false; };
};

class Test1Parameter : public Parameter { 	/* User value 7999 -> $7FFFFF in DSP */

public:

	Test1Parameter(char *n, char *sn, int min = 0, int max = 7999);

	virtual int 	translate(int value, int source);
};

class Test2Parameter : public Parameter {	/* User range 0..32 -> -524288..524288 in DSP */

public:

	Test2Parameter(char *n, char *sn, int min = 0, int max = 31);

	virtual int 	translate(int value, int source);
};

class PitchParameter : public Parameter {

public:

	PitchParameter(char *n, char *sn);

	virtual int 	translate(int value, int source);
	virtual char	*baseValStr(int value, int &strLen);
	virtual bool	isStoppedAt1000() { return false; };
};

class CutoffParameter : public Parameter {

public:

	CutoffParameter(char *n, char *sn, int min = -8000, int max = 7999);

	virtual int 	translate(int value, int source);
	virtual int	mTranslate(int depth);
	virtual char	*baseValStr(int value, int &strLen);
	virtual bool	isStoppedAt1000() { return false; };
};

class LevelParameter : public Parameter {

public:	

	LevelParameter(char *n, char *sn, int min = 0, int max = 7999);
	
	virtual int 	translate(int v, int source);
	virtual int 	mTranslate(int depth);
	virtual bool	usesPoPsMatrix() { return true; };
};

class VolumeParameter : public LevelParameter {

public:	

	VolumeParameter(char *n, char *sn, int min = 0, int max = 7999);
	virtual bool isMeteredParameter() { return true; };
};

class StringParameter : public SwitchParameter {

protected:

	char	**paraString;
	int	*paraLength;
	int	addIndex;
	void	addString(char*);

public:

	StringParameter(char*, char*, int n);
	virtual ~StringParameter();
	
	virtual char	*getValStr(int value, int source, int &strLen);	
	virtual bool	isMutable() { return FALSE; } ;	
};

class OnOffParameter : public StringParameter {

public:
	OnOffParameter(char *n, char *sn);
	virtual bool isMutable() { return TRUE; } ;	
};

class VoiceAssignParameter : public OnOffParameter {

public:
	VoiceAssignParameter(char *n, char *sn);
	virtual bool isDspParameter() { return false; };
};

class WavetypeParameter : public StringParameter {

public:
	WavetypeParameter(char *n, char *sn);
};

class SourceParameter : public StringParameter {

public:
	SourceParameter(char *n, char *sn);
};

class LFOWaveformParameter : public StringParameter {

public:
	LFOWaveformParameter(char *n, char *sn);
};

class vLFOSyncParameter : public StringParameter {

public:
	vLFOSyncParameter(char *n, char *sn);
	virtual bool isMutable() { return TRUE; } ;	
};

class sLFOSyncParameter : public StringParameter {

public:
	sLFOSyncParameter(char *n, char *sn);
	virtual bool isMutable() { return TRUE; } ;	
};

//----------------------------------------------------------------------------
//
// 			      Pages
//
//----------------------------------------------------------------------------
 
class Page : public ParaName {

protected:
	
	Parameter* 	parameter[MAX_PARAS_PER_PAGE];	// The array of parameters
	int		currentParameterIndex;
	char		potLUT[128];
	
	static int 	nPages;		// Counts no. of Page objects
	int		nParas;		// No. of Parameters in page
	int		groupIndex;	// Index of group which contains this page
	int		pageIndex;	// Index of this page within the group
	
public:

	Page(char *longName, char *shortName);
	
	inline int 	getnParas() { return nParas; };

	Parameter*	getParameterByIndex(int x);
	Parameter*	getCurrentParameter();
	int	  	getCurrentParameterIndex();
	int	  	setCurrentParameterByIndex(int x);
	bool	  	setCurrentParameterByPot(int p);
	bool		parameterInc(int x);

	static void countZero() { nPages = 0; };
	static int getCount() { return nPages; };
	
	inline int	getGroupIndex() { return groupIndex; };
	inline int	getPageIndex() { return pageIndex; };

	/* Group and page indeces are set by the parent group when it is instantiated */
	inline void	setIndeces(int gi, int pi) { groupIndex = gi; pageIndex = pi; };
};

class OscillatorPage : public Page {

public:	OscillatorPage(char* n, char* sn);
	
};

class FilterPage : public Page {

public:	FilterPage(char* n, char* sn);
	
};

class NoiseGenPage : public Page {

public:	NoiseGenPage(char* n, char* sn);

};

class ExtInputPage : public Page {

public:	ExtInputPage(char* n, char* sn);

};

class RingModPage : public Page {

public:	RingModPage(char* n, char* sn);

};

class vLFOPage : public Page {

public:	vLFOPage(char* n, char* sn);
	
};

class sLFOPage : public Page {

public:	sLFOPage(char* n, char* sn);
	
};

class ENVPage : public Page {

public:	ENVPage(char* n, char* sn); 
	
};

class OutputPage : public Page {

public:	OutputPage(char* n, char* sn);
	
};

//----------------------------------------------------------------------------
//
// 			      Groups
//
//----------------------------------------------------------------------------
 
class Group {

protected:
	
	Page* 		page[MAX_PAGES_PER_GROUP];	
	int		currentPageIndex;		
	int		nPages;		

public:

	Group();
	
	Page* 		getPageByIndex(int x);
	Page* 		getCurrentPage();
	int		getCurrentPageIndex();
	int		setCurrentPageByIndex(int x);
	bool		pageInc(int x);
	
	inline int 	getnPages() { return nPages; };
	Page*		addPage(Page* page, int groupIndex);
};

//----------------------------------------------------------------------------
//
// 			   ParaStructure
//
//----------------------------------------------------------------------------

class ParaStructure {

	Group*		group[MAX_GROUPS];			// The array of Page Groups
	Page*		page[MAX_GROUPS * MAX_PAGES_PER_GROUP];	// A flat array of Parameter Pages

	char		pagePotLUT[128];	// Maps pot values to Pages (page indeces)
	char		matrixPotLUT[128];	// Maps pot values to matrix sources
	
	int		currentGroupIndex;		// For editing
	int		currentSource;
	
	Parameter**	dspPara;		// Parameter list in order used by dsp	
	Parameter**	userPara;		// Parameter list in order used by interface
	
	int	nGroups;		// Number of Groups in parameter structure
	int	nPages;			// Total number of Pages in all groups
	int	nParas;			// Total number of Parameter objects in all Pages
	int	nDspParas;		// Total number of parameter objects relevant to DSP
	int	nMatrixSources;		// Number of matrix sources
	int	nValues;		// Number of values, ie nParas + nParas * nMatrixSources
	int 	nameLength;		// Length of name
	int	tuningLength;		// Length of tuning data
	int	tuningNameLength;	// Length of tuning name
	
	Hash*	hashTable;		// Access parameters by name (slow to initialise - dev only)
	void	makeHash();		// Called when hashTable is first used

	void	makeDspParaList();	// Creates list of parameters in order used by dsp, 
					// also updating each parameter's dspParaIndex value

	void	makeUserParaList();	// Creates list of parameters in order used in interface
	
	void	generateDspCodeLine(Page*, Parameter*);		// Used during development
	
public:

	bool	setCurrentGroupPageParaByIndex(int index); 	
	
	Parameter* getDspPara(int);		// Used in all dsp parameter communication
	Parameter* getParaByIndex(int);		// arg is the base parameter number (ie 0..89)
	Parameter* getParaByName(char*);	// Use combination of short Page and Para names
						// ie "LFO1 wave", "FLT bal", "Portamnto"
	int	setCurrentGroupByIndex(int x);
	Group*	getCurrentGroup();
	bool	groupInc(int x);
	
	int	setCurrentSource(int s);
	int	getCurrentSource();
	bool	sourceInc(int x);	
	
	Page*	getPageByIndex(int x);		// (uses flat page index, without groups)
	char*	getSourceName(int);
	
	bool	setCurrentSourceByPot(int p);	// For selecting paras by pot
	bool	setCurrentPageByPot(int p);

	inline int getnParas() { return nParas; };
	inline int getnDspParas() { return nDspParas; };
	inline int getnPages() { return nPages; };
	inline int getnGroups() { return nGroups; };
	inline int getnMatrixSources() { return nMatrixSources; };
	inline int getnameLength() { return nameLength; };
	inline int getnValues() { return nValues; };
	inline int gettuningLength() { return tuningLength; };
	inline int gettuningNameLength() { return tuningNameLength; };
	
	ParaStructure();
};

//----------------------------------------------------------------------------
//
// 				EntitySound
//
//----------------------------------------------------------------------------

class EntitySound {

	ParaStructure *paras;	// Pointer to the ParaStructure object passed in constructor arg.
	
	int	*value;		// Values user sees on screen, and that are stored in flash
	int	*oldval;	// Previous user-visible values (to check for changed sound)
	bool	*mute;		// If true, has the effect of setting value to zero

	int	number;		// Sound number
	char	*name;		// Sound name
	char	*oldName;

	int	*tuning;	// Tuning LUT, 128 offsets from middle C in  cent units
	int	tuningNumber;	// From midi tuning std (unused)
	char	*tuningName;

	char	*pot;		// Control pot values (so when sound is loaded, it sounds like
				// last time it was used, because pot values have been restored)

	int	nChangedParas;	// Counts number of changed paras; if zero, sound is unchanged.
	int	nMatrixLinks;	// Counts number of non-zero matrix links
	
	bool	available;	// Relates to pool of sound objects, indicates whether a sound
				// is available for immediate use, or not (ie, is being used by 
				// synth, awaiting use by synth, or being used for loading)
	
	int 	poolIndex;	// Index of sound in sound pool (see EntitySoundManager.cpp)
public:

	unsigned int	checksum() const;
	int		compress(void *buffer);
	void		deCompress(void *buffer);
	void		setTuning(void *buffer);
	
	bool 		findNextMatrixDestination(int mSrc, Parameter*);
	int		findNextMatrixSource(int mSrc, Parameter*);
	
	void	clearSound();
	void	loadRandomValues();
	void 	loadDefaultValues(int = 0);

	int 	getValue(Parameter* parameter, int source = -1);
	int 	getTranslatedValue(Parameter* parameter, int source = -1);
	int 	gettuning(int note);
		
	Parameter *getParameterByIndex(int paraIndex);
	
	Status	incValue(Parameter* parameter, int incr, int source = -1);
	Status	changeValue(Parameter *parameter, int newValue, int source = -1);
	Status	setMute(Parameter* parameter, bool mute, int source = -1);
	Status	toggleMute(Parameter* parameter, int source = -1);
	
	bool	loadValue(char *paraname, int intendedValue, int source = -1);
	void	loadValue(Parameter *parameter, int intendedValue, int source = -1);

	char	*getValStr(Parameter* parameter, int source, int &strlen);

	bool	getMute(Parameter* parameter, int source = -1);
	void	updatenChangedParas(int, int, int);
	
	char	getBullet() const;
	void	setName(char*);
	
	static 	char *getNumberStr(int);				// Utility: get string for any sound number
	inline	char *getNumberStr() { return getNumberStr(number); };	// Object: get string for this sound's number	
	
	inline	char *getName() { return name; };
	inline 	static char *getDefaultName() { return "            "; };
	inline 	bool nameChanged() { return (bool) strncmp(oldName, name, paras->getnameLength()); };

	inline	bool isAvailable() { return available; };
	inline 	void makeAvailable() { available = true; };
	inline	void makeUnavailable() { available = false; };	
	
	inline 	int getNumber() { return number; };			// Unformatted, range 0..large number
	inline	void setNumber(int n) { number = n; };
	
	inline	void setAllParasChanged() { nChangedParas = 99999; };	// For when sound loaded from midi
	inline	int getnChangedParas() const { return nChangedParas; };
	inline 	int getnMatrixLinks() const { return nMatrixLinks; };
	
	inline	bool isModified() const { return (bool) nChangedParas; };
	
	inline	int setPot(int p, int v) { return pot[p] = v; };
	inline	int getPot(int p) { return pot[p]; };

	inline	void setPoolIndex(int x) { poolIndex = x; };
	inline 	int getPoolIndex() { return poolIndex; };
	
	EntitySound(ParaStructure*);
	~EntitySound();
	
};







	










	



#endif /* __ENTITYSOUND_H__ */
