/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntitySoundManager.cpp							*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/
  
#include "EntitySoundManager.h"

//#define DEBUG_SOUND_MANAGER

bool EntitySoundManager::bgLoadSound(int n) {

	newNumber = n;
	
	rtems_semaphore_obtain(semaphore, RTEMS_WAIT, RTEMS_NO_TIMEOUT);
	rtems_task_suspend(task_id);
	rtems_semaphore_release(semaphore);

	status = rtems_task_restart(task_id, (rtems_task_argument) this);
	if (status != RTEMS_SUCCESSFUL) { TRACE("ERROR: cannot start background loader.\n"); return false; }	
	return true;
}

bool EntitySoundManager::loading() {

	 return rtems_task_is_suspended(task_id) == RTEMS_ALREADY_SUSPENDED ? false : true;
}

void EntitySoundManager::Run() { // Background loading

	// First time round do queue id initialisation and suspend
	
	if(!taskFlag) {	taskFlag = true; rtems_task_suspend(RTEMS_SELF); }
	
	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	 * N.B. It is VERY important that this function should NOT change the flash
	 * state as this task will be halted and restarted at random. N.B. it can
	 * change the flash version access state, as this is only a file state.
	 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

	// This is quite subtle, if your familiarity with the code has faded. Once
	// a sound file has been loaded, it is designated the currentSound. However,
	// it is still waiting for the synth to actually use it. There are conflicting
	// requirements: You must re-use the sound pool objects, but you can't load into
	// the current sound in case it suddenly gets used by the synth. This conflict 
	// is resolved by ensuring that the nextSound is not the currentSound. If you 
	// get to the end of loading the nextSound you can make it the new currentSound 
	// before the previous one gets used by the synth, thus re-using the sound objects 
	// in the pool without the conflict. The currentSound as seen by the sound manager
	// is not necessarily the current sound that is being heard - that is an internal 
	// matter for the synth. The two are only the same, after a note has been played.

	nextSound = getSoundFromPool();

	// Quickly load sound name

	char *n = flash->snoopName(newNumber);
	strcpy(newName, n);	
	sendMessage(ctrq, NEW_SOUND_NAME_READY);

	rtems_task_wake_after(50);	// Pause in case loading is unnecessary
	
	// Full sound loading
	
	if(!flash->load(nextSound, newNumber)) TRACE("ERROR - could not load sound\n");
		
	sendMessage(ctrq, NEW_SOUND_READY);

	// Finalisation - here it will be considered that, although the new sound won't
	// be heard until the next midi note-on is received, the new sound is the 
	// "current sound" and its parameters can be edited, even if the
	// old sound is still heard.

	activate(nextSound);
	
	rtems_task_suspend(RTEMS_SELF);
}

bool EntitySoundManager::loadSound(int n) { // Like background loading but without messaging

	while(loading()) rtems_task_wake_after(RTEMS_YIELD_PROCESSOR);

	nextSound = getSoundFromPool();
	if(!flash->load(nextSound, n)) { TRACE("ERROR: could not load sound\n"); return false; }
	activate(nextSound);
	
	return true;
}

int EntitySoundManager::loadMidiFromSound(int &soundNumber, void *buffer) {

	soundNumber = currentSound->getNumber();
	return currentSound->compress(buffer);
}

bool EntitySoundManager::loadSoundFromMidi(int soundNumber, void *buffer) {

	while(loading()) rtems_task_wake_after(RTEMS_YIELD_PROCESSOR);

	nextSound = getSoundFromPool();
	nextSound->deCompress(buffer);
	nextSound->setNumber(soundNumber);

	nextSound->setAllParasChanged(); // So user can see an update has occured
	
	activate(nextSound);

	return true;
}

bool EntitySoundManager::loadTuningFromMidi(void *buffer) {

	currentSound->setTuning(buffer);
	nextSound->setAllParasChanged(); 

	return true;
}

EntitySound *EntitySoundManager::getSoundFromPool() {

	/* Find an available sound object in pool (reuse order unimportant) */
	
	EntitySound *nSound = 0;
	int countdown = 1000;
	int available = 0;
	
	while(!available && countdown > 0) {
	
		for(int x = 0; x < N_SOUND_OBJECTS; x++) {
		
			EntitySound *s = sound[x];
		
			if(	s->isAvailable() && 
				s != currentSound && 
				s != editedSound1 &&
				s != editedSound2) {
		
				nSound = sound[x];
				available++;

				#ifdef DEBUG_SOUND_MANAGER 
				TRACE("Sound %i: available\n", x);
				} else  {		
				TRACE("Sound %i: unavailable\n", x);
				#endif
			}
		}

		/* If none available, give synth time to free sounds */		
		if(!available) { rtems_task_wake_after(10); countdown -= 10; }
	}

	#ifdef DEBUG_SOUND_MANAGER 
	TRACE("Available sound objects = %i\n", available);
	#endif

	if(countdown <= 0) TRACE("ERROR: Timeout for sound pool.\n");
	
	return nSound;
}

void EntitySoundManager::activate(EntitySound* snd) {

	int nBytes = 0;

	while(!nBytes) {

		rtems_semaphore_obtain(semaphore, RTEMS_WAIT, RTEMS_NO_TIMEOUT);

		/* These steps must be executed atomically. This is not because they cannot be interrupted
		 * but because they MUST be completed. Unless pre-emption is disabled, it is possible
		 * (though very rare) for snd->makeUnavailable and sendMessage(sstq...) to be separated,
		 * resulting in permanent loss of a sound from the sound pool, or, if ordered differently, 
		 * re-using a sound that is being transfered to the dsp. This is because background loading 
		 * can be restarted at any time, unless pre-emption is temporarily turned off at this point. */
		
		nBytes = sendMessage(synq, NEW_SOUND_READY);
		
		if(nBytes) { // if SYNQ sucessfully written, okay to proceed with the other atomic steps
			
			/* Since there are only a few sounds in the pool, SSTQ should never get too full */
			if(!sendMessage(sstq, &snd, 4)) TRACE("FATAL ERROR: SSTQ not written\n");
		
			snd->makeUnavailable();	

			currentSound = snd;
		}

		rtems_semaphore_release(semaphore);
		
		if(!nBytes) {
			
			TRACE("SYNQ full, waiting to clear\n");
			rtems_task_wake_after(10); // Give SYNQ or SSTQ time to clear
		}
	}
}

void EntitySoundManager::update() {

	/* Worth describing this rather cryptic function: editing modifies
	 * a sound object, but the editedSound ptr doesn't get updated 
	 * because editing doesn't know about sound management - it just
	 * works on a sound. When sound management is used, therefore,
	 * before anything is done, this function, update(), must be
	 * called to syncronise the state. */
	
	if(currentSound->isModified()) {
		
		// If previously edited version of current sound exists, use that pointer
		
		if(currentSound->getNumber() == editedSound1->getNumber()) {

			editedSound1 = currentSound;
			editedSoundLRU = &editedSound2;
		}

		else if(currentSound->getNumber() == editedSound2->getNumber()) {

			editedSound2 = currentSound;
			editedSoundLRU = &editedSound1;
		}

		// Use least-recently used edited sound pointer
		
		else if(editedSoundLRU == &editedSound1) {

			editedSound1 = currentSound;
			editedSoundLRU = &editedSound2;
			
			if(editedSound2 == currentSound) editedSound2 = 0;
		}
		
		else if(editedSoundLRU == &editedSound2) {

			editedSound2 = currentSound;
			editedSoundLRU = &editedSound1;
			
			if(editedSound1 == currentSound) editedSound1 = 0;
		}
	}
	
	/* And this is subtle too: when a modified sound has been saved, 
	 * editedSound still points to it, even though after saving it
	 * is no longer a modified sound. The editedSound ptr should 
	 * be nulled, otherwise the version navigation logic won't 
	 * work. The nulling is done here, rather than provide 
	 * editing with sound manager access. */

	else {
		if(editedSound1 == currentSound) {
			
			editedSound1 = 0;
			editedSoundLRU = &editedSound1;

			return;
		}
			
		else if(editedSound2 == currentSound) {
			
			editedSound2 = 0;
			editedSoundLRU = &editedSound2;

			return;
		}
	}
}

bool EntitySoundManager::versionUp() {

	while(loading()) rtems_task_wake_after(50);

	// Cannot go higher if current sound is edited

	if(currentSound == editedSound1) return false;
	if(currentSound == editedSound2) return false;
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// If at most recent stored version, try edited sound

	if(!flash->getversion()) { // Already at most recent stored sound ?

		// Cannot go higher if there is no edited sound available
		
		if(!editedSound1 && !editedSound2) return false; 

		// Go to edited sound with same number

		if(editedSound1->getNumber() == currentSound->getNumber()) {

			activate(editedSound1);
			editedSoundLRU = &editedSound2;
			
			return true;		
		}

		else if(editedSound2->getNumber() == currentSound->getNumber()) {
			
			activate(editedSound2);
			editedSoundLRU = &editedSound1;

			return true;
		}
				
		return false;	// if none of the edited sounds matches 
	}
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Current sound is an old version, so load the next
	
	nextSound = getSoundFromPool();

	bool changed = flash->loadNextVersion(nextSound);
	if(changed) activate(nextSound);

	return changed;
}

bool EntitySoundManager::versionDown() {

	while(loading()) rtems_task_wake_after(50);

	int sn = currentSound->getNumber();

	// - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// if current sound is edited, load the unedited version

	if(currentSound == editedSound1 || currentSound == editedSound2) { 
	
		nextSound = getSoundFromPool();

		if(!flash->getnVersions()) flash->load(nextSound, sn);
		else flash->loadVersion(nextSound, 0);

		activate(nextSound);
		return true;	
	}

	// versions of current sound are ready for access?
	if(!flash->getnVersions()) flash->initVersionAccess(sn);
	
	// already at one and only (and unmodified) version of sound?
	if(flash->getnVersions() == 1) return false; 

	// - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Multiple versions exist, so load the previous version 
	
	nextSound = getSoundFromPool();

	bool changed = flash->loadPreviousVersion(nextSound);
	if(changed) activate(nextSound);

	return changed;
}

int EntitySoundManager::getversion() {
	
	if(currentSound->isModified()) return -1;
	return flash->getversion(); // 0..n
}

int EntitySoundManager::getnVersions() {

	return flash->getnVersions();
}

EntitySoundManager::EntitySoundManager(EntityFlash *flash_, ParaStructure *paras_) {

	flash = flash_;
	paras = paras_;
	
	taskFlag = false;

	for(int x = 0; x < N_SOUND_OBJECTS; x++) {

		sound[x] = new EntitySound(paras);
		sound[x]->makeAvailable();
		sound[x]->setPoolIndex(x);
	}

	ctrq = getQueueId("CTRQ");
	synq = getQueueId("SYNQ");
	sstq = getQueueId("SSTQ");
	
	nextSound = 0;

	currentSound = 0;
	editedSound1 = 0;
	editedSound2 = 0;

	editedSoundLRU = &editedSound1;
	
	status = rtems_semaphore_create(rtemsName("SMNS"), 1, RTEMS_FIFO, 50, &semaphore);
}

EntitySoundManager::~EntitySoundManager() {

	for(int x = 0; x < N_SOUND_OBJECTS; x++) {

		delete sound[x];
	}
}

