/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntitySoundManager.h							*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/

#ifndef __ENTITYSOUNDMANAGER_H__
#define __ENTITYSOUNDMANAGER_H__

#include <rtems.h>
#include <chameleon.h>
#include <string.h>

#include "EntitySound.h"
#include "EntityFlash.h"
#include "EntityProcess.h"
#include "EntityMessages.h"

#define N_SOUND_OBJECTS 6

/* The sound manager operates on a pool of a few sound objects,that are rotated to 
 * facilitate background loading and updating.Once a sound is ready, it can be 
 * activated on the next note. */

/* This class has two kinds of operation: background loading, which is a task, 
 * and sound managing, which comprises a number of utilities that can be invoked 
 * by another task, including the function to initiate background loading. */

class EntitySoundManager : public EntityProcess {

	EntityFlash*	flash;				// Flash filing system for loading sounds
	ParaStructure*	paras;				// Parameter structure for sound

	EntitySound*	sound[N_SOUND_OBJECTS];		// Sound object pool (objects instantiated by constructor)
	EntitySound*	currentSound;			// Pointer to current (most recent fully loaded) sound
	EntitySound*	editedSound1;			// Pointer to a modified sound (possibly = currentSound)
	EntitySound*	editedSound2;			// Pointer to a modified sound (possibly = currentSound)
	EntitySound*	nextSound;			// Pointer to next sound to be used for loading
	EntitySound**	editedSoundLRU;			// Pointer to least-recently-used of editedSound1 or 2
	
	char		newName[SOUND_NAME_LENGTH];	// Name of sound being loaded (not necessarily currentSound's name)
	int		newNumber;			// Sound number of sound to load (for passing para to bg loader)
	
	bool		taskFlag;			// Enables task to go dormant on first starting

	rtems_id 	ctrq;				// Queue id for signalling the control process
	rtems_id 	synq;				// Queue id for signalling the synth
	rtems_id 	sstq;				// Queue id for sound stack used by synth

	rtems_id	semaphore;			// Semaphore used for atomisation of activate()
	
	EntitySound*	getSoundFromPool();
	void		activate(EntitySound*);

public:

	EntitySoundManager(EntityFlash*, ParaStructure*);
	~EntitySoundManager();
	
	void 		Run();			// Background task

	bool 		bgLoadSound(int);	// Initiate multi-stage background loading of sound indexed by arg
	bool		loading();		// Indicates if background loading is taking place
	bool		loadSound(int);		// Simple function call to load sound (no background loading)
	
	int		loadMidiFromSound(int&, void*);	// Load current sound to midi sysex buffer
	bool		loadSoundFromMidi(int, void*); 	// Load from midi sysex buffer
	bool		loadTuningFromMidi(void*);	// Load new tuning into current sound
	
	bool		versionUp();		// Get newer or edited version of sound
	bool		versionDown();		// Get previous version of sound

	int		getversion();		// Get version no. of current sound (0 = most recent)
	int		getnVersions();		// Get no. versions of current sound (0 = versions need initialising)
	
	void		update();		// Update state of sound manager (ie, on entry to EntitySoundMode)
	
	inline EntitySound*	getSound() { return currentSound; };
	inline int		getNumber() { return currentSound->getNumber(); };
	inline char		getBullet() { return currentSound->getBullet(); };
	inline char* 		getNameStr() { return currentSound->getName(); };
	inline char* 		getNewNameStr() { return newName; };
	inline char* 		getNumberStr(int n) { return EntitySound::getNumberStr(n); }
	inline int 		getMaxNumber() { return flash->getnSounds() - 1; };
	inline int 		getMaxWritableNumber() { return flash->getnWritableSounds() - 1; };
	inline int		getBankSize() { return SOUND_BANK_SIZE; };
};










#endif /* __ENTITYSOUNDMANAGER_H__ */
