/********************************************************************************
 * U T I L S									*
 ********************************************************************************
 * File: utils.h								*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/

#ifndef __UTILS_H__
#define __UTILS_H__

#define MAGIC1 12345678			// The default seed for any rng
#define MAGIC2 89123456			// For fixing bad slices in r250::validate

//--------------------------------------------------------------------------------
// Here we have a lesser rng. This can be used to seed an r250 in preference to 
// defining each of the 250 chars of the r250 individually.

class rshift {

public:
        unsigned int state;     		// 32 bits of state
        int rng();    				// Generate 32 bit rng from high bits of two rn() outputs
	rshift(unsigned int seed = MAGIC1);	// Constructor

private:
	int rn();				// Generate 32 bit rng with poor quality low bits

};

//----------------------------------------------------------------------------------------
// Here we have the r250 random number algorithm.

// r250 luts for fast rng - required in r250 declaration

class r250 {
	
public:
        int state[250];       							// data defining r250 state
       
	int rng() { return (state[si=incr[si]] ^= state[sj=incr[sj]]); } 	// generate next number
	
        int gnr() { state[si] ^= state[sj];					// generate previous number
		    sj=decr[sj]; 
		    return state[si=decr[si]]; }				
		
	int getpos(); 				// get position of si in state
	void setpos(int n);			// set position of si and sj
	
	void validate();        	     	// used to ensure bit columns of state are all different
	
	void seed(int seed);			// manually sow state from single number seed
	void seed(int* seed);			// manually sow state from array of numbers

	void spin(int n);			// iterate n times
	
	r250(int seed = MAGIC1);		// constructor

private:
	
	unsigned int  mask[32];			// Bit mask lut, for addressing bits { 0001, 0010, 0100...}
	unsigned char incr[250];   		// Used to do fast wraparound of i and j in r250
	unsigned char decr[250];   		// Same again, but when going backwards
	
	unsigned char si, sj;             			// Pointers into state, sj = si + 103 
	void fix_slice(int slice[8][32], int j, rshift &rx);	// Used in validate() to conform state
};                 

//-------------------------------------------------------------------------
// A simple binary tree for text

class BTree {

	BTree	*low;
	BTree	*high;
	char	*key;
	void	*data;

	int	compare(char*);

public:		

	BTree(char*, void*);
	~BTree();

	void	*lookup(char*);
	bool	add(char*, void*);
	int	size();
};
	
//---------------------------------------------------------------
// A simple hash for text lookup

class Hash {

	BTree	**table;

	int	tableSize;
	int	rTable[256];
	
	int	hashFunction(char*);
	
public:

	Hash(int hTableSize = 4096);
	~Hash();

	bool	add(char*, void*);
	void	*lookup(char*);

	void	getStats(int &direct, int &total);
};

//---------------------------------------------------------------------------------------------------------
// Quicksort 
//
// The way this works is that you have a list of entities, elements, each of which is typically 
// represented as a number (which is elsewhere used probably to index an array of element objects), 
// plus you have a parallel list of their scores. Unlike an alphabetic sort, where the elements are
// self-scoring, both the scores and their list elements are swapped. 

template <class L, class S> 

class QuickSort {
	
public:
	QuickSort(L *l, S *s, int n=1);		// Instantiate with list of elements and list of their scores
						// and also the number of elements	
	void sort();
	void sort(int n);	
	void sort(S *&s);
	
private:
	void sort(int low, int high);		// the recursive call
	void swap(int a, int b);
	
	L *list;				// List of entities
	S *score;				// List of their scores
	
	int listsize;				// number of 
};

template <class L, class S> QuickSort<L,S>::QuickSort(L *l, S *s, int n) {
       
	list = l;
	score = s;
	listsize = n;
}	

template <class L, class S> void QuickSort<L,S>::sort() {
	
	sort(0, listsize-1);
}

template <class L, class S> void QuickSort<L,S>::sort(int n) {
	
	sort(0, n-1);
}

template <class L, class S> void QuickSort<L,S>::sort(S *&s) {
	score = s;
	sort(0, listsize-1);
}

template <class L, class S> void QuickSort<L,S>::sort(int low, int high) {
	
	/* Initialize pointers */

	int	top = low;
	int	bottom = high - 1;
	L	part_index;
	S	part_value;

	// Do nothing if low >= high
	
	if (low < high) {
    		
		/* Check if elements are sequential */
		
		if (high == (low + 1)) {
			if (score[low] < score[high])
				swap(high, low);
		}
	
		else {
		
		/* Choose a partition element and swap it with the last value in the array */

			part_index = (int)((low + high) / 2);
			part_value = score[part_index];
			swap(high, part_index);

			do {	/* Increment the top pointer */
				
				while ((score[top] >= part_value) && (top <= bottom)) top++;
				
				/* Decrement the bottom pointer */
				
				while ((score[bottom] < part_value) && (top <= bottom)) bottom--;
				
				/* Swap elements if pointers have not met */
				
				if (top < bottom) swap(top, bottom);
				
			} while (top < bottom);
			
			/* Put the partition element back where it belongs */
			
			swap(top, high);

			/* Recursive calls */
			
			sort(low, top - 1);
			sort(top + 1, high);
		}
	}
}

template <class L, class S> void QuickSort<L,S>::swap(int a, int b) {		// Swap list elements

	S s; s = score[a]; score[a] = score[b]; score[b] = s;
	L l; l = list[a]; list[a] = list[b]; list[b] = l;
}

//-------------------------------------------------------------------------
// CRC32 Checksum Class

class CRC32 {

	unsigned int crcLUT[256];
	unsigned int crc;

public:

	CRC32();
	
	unsigned int 	checksum(void *buffer, unsigned int nBytes);
	unsigned int 	checksum() { return crc; }
	void 		init() { crc = 0xffffffff; }
	void 		add(unsigned char c) { crc = (crc >> 8) ^ crcLUT[(crc ^ c) & 0xff]; }
};

//------------------------------------------------------------------
// Bitstream Class

class BitStream {

	unsigned char	*ptr;
	int		mask;
	int		byte;
		
public:

	void		init(void *buffer);	
	int		get(int nBits);
	void		put(int value, int nBits);
	int		getnBytes();
};

//------------------------------------------------------------------
// Fast integer to string conversion function
//
// This is a right justified sign plus digits number-to-string 
// converter. There is no checking of the input value.

char *i2s(int);

//---------------------------------------------------------------------------------------------------------
// Create potentiometer value map
//
// When a 128 step potentiometer value is mapped onto around twenty or less discrete values, the 
// dead regions at the end of travel can be utilised to give a greater (on average) angular 
// displacement per parameter.
//
// The pot LUT may be mapped to actual values, such as integers, or any array of objects.
// 
// *potLUT must point to an array of 128 ints or object pointers on entry, 
// and this will be updated from the contents of the objectArray.

template <class T> void initPotLUT(int nObjects, T *potLUT, T *objectArray = 0) {

	const int virtualPotRange = 139 << 8; // (* 256 for precision)
	const int potOffset = (virtualPotRange-(128 << 8)) >> 1;

	int splodge = virtualPotRange / nObjects; 

	if(objectArray) {

		for(int potValue = 0; potValue < 128; potValue++) {
	
			int index = ((potValue<<8)+potOffset) / splodge;
			potLUT[potValue] = objectArray[index];
		}
	
	} else {
			
		for(int potValue = 0; potValue < 128; potValue++) {
	
			int index = ((potValue<<8)+potOffset) / splodge;
			potLUT[potValue] = (T) index;
		}	
	}
}


	



















#endif /* __UTILS_H__ */
